from typing import Any


def run_experiment(**kwargs: Any) -> None:
    # TODO: Implement run_experiment
    print("Your experiment results are available at: https://app.galileo.ai/experiments/results")
